# -*- encoding: utf-8 -*-
"""
SALLY
sally.core.serving module

Endpoint service
"""
import os
from typing import List

import falcon
from base64 import urlsafe_b64encode as encodeB64

from hio.base import doing, Doer
from hio.core import http
from hio.help import decking
from keri.app import indirecting, storing, notifying, habbing, oobiing
from keri.app.cli.commands import incept
from keri.app.habbing import Habery, Hab
from keri.core import routing, eventing, parsing
from keri.end import ending
from keri.peer import exchanging
from keri.vdr import viring, verifying
from keri.vdr.eventing import Tevery
from keri.vc import protocoling

from sally import ogler, log_name
from sally.core import handling, basing, monitoring, httping
from sally.core.credentials import TeveryCuery
from sally.core.verifying import VerificationAgent

logger = ogler.getLogger(log_name)

def incept_if_new(hby: Habery, alias: str, incept_args: dict):
    """
    Create an identifier with the specified alias if it does not exist.
    This is run in a separate scheduler (Doist) prior to starting Sally so that the identifier can
    exist and have all Habery configuration OOBIs resolved.

    Warning: This only supports no witness, single sig identifiers as it is assumed it will be used
        with the direct-mode verification agent.
    """
    if incept_args is None:
        incept_args = {}
    hab = hby.habByName(name=alias)
    if hab is None:
        # Incept the new Hab
        if incept_args["incept_file"] is None:
            raise ValueError("incept file by arg --incept-file is required to create a new identifier")
        logger.info(f"Making new hab {alias} in Habery {incept_args.get('name', '')}")
        hab = hby.makeHab(name=alias, **inception_config(**incept_args))
    else:
        logger.info(f"Hab '{alias}' already exists, using...")
    return hab

def run_doers(doers: List[Doer], expire=0.0):
    """Runs the verifier in a Doist until the specified expire time (in seconds)."""
    tock = 0.03125
    doist = doing.Doist(limit=expire, tock=tock, real=True)
    doist.do(doers=doers)


def setupDoers(
        hby: Habery, hab: Hab, alias: str, http_port: int, hook: str, auth: str,
        timeout: int = 10, retry: int = 3, direct: bool = True, incept_args: dict = None):
    """
    Parameters:
        hby (Habery): identifier database environment
        alias (str): alias of the identifier representing this agent
        httpPort (int): external port to listen on for HTTP messages
        hook (str): URL of external web hook to notify of credential issuance and revocations
        auth (str): alias or AID of external authority for contacts and credentials
        timeout (int): escrow timeout (in minutes) for events not delivered to upstream web hook
        retry (int): retry delay (in seconds) for failed web hook attempts
        direct (bool): listen for direct-mode messages on HTTP port or use indirect-mode mailbox
        incept_args (dict): arguments for incepting Sally's identifier if it does not exist
    """
    hbyDoer = habbing.HaberyDoer(habery=hby)
    obl = oobiing.Oobiery(hby=hby)

    doers = [hbyDoer, *obl.doers]
    doers += setup(hby, hab, alias=alias, httpPort=http_port, hook=hook, auth=auth, timeout=timeout,
                   retry=retry, direct=direct)
    return doers

def setup(hby: Habery, hab: Hab, alias: str, httpPort: int, hook: str, auth: str, timeout=10,
          retry=3, direct=True):
    """
    Setup components, HTTP endpoints, and MailboxDirector working with witnesses to receive events.

    Parameters:
        hby (Habery): identifier database environment
        alias (str): alias of the identifier representing this agent
        httpPort (int): external port to listen on for HTTP messages
        hook (str): URL of external web hook to notify of credential issuance and revocations
        auth (str): alias or AID of external authority for contacts and credentials
        timeout (int): escrow timeout (in minutes) for events not delivered to upstream web hook
        retry (int): retry delay (in seconds) for failed web hook attempts
        direct (bool): listen for direct-mode messages on HTTP port or use indirect-mode mailbox
    """
    cues = decking.Deck()
    if hab is None:
        raise ValueError(f"Identifier with alias '{alias}' does not exist. It must be created before starting Sally.")

    logger.info(f"Using hab {hab.name}:{hab.pre}")
    logger.info(f"\tCESR Qualifed Base64 Public Key:  {hab.kever.serder.verfers[0].qb64}")
    logger.info(f"\tPlain Base64 Public Key:          {encodeB64(hab.kever.serder.verfers[0].raw).decode('utf-8')}")

    # HTTP Server
    app = falcon.App(middleware=httping.cors_middleware())
    server = http.Server(port=httpPort, app=app)
    httpServerDoer = http.ServerDoer(server=server)

    # KEL, ACDC, exchange message, and reply message components
    reger = viring.Reger(name=hab.name, db=hab.db, temp=False)
    verifier = verifying.Verifier(hby=hby, reger=reger)

    mbx = storing.Mailboxer(name=hby.name)
    exc = exchanging.Exchanger(hby=hby, handlers=[])
    rep = storing.Respondant(hby=hby, mbx=mbx)

    cdb = basing.CueBaser(name=hby.name)
    clear_escrows(cdb)

    rvy = routing.Revery(db=hby.db)
    notifier = notifying.Notifier(hby=hby)
    # writes notifications for received IPEX grant exn messages
    protocoling.loadHandlers(hby=hby, exc=exc, notifier=notifier)

    kvy = eventing.Kevery(db=hby.db, lax=True, local=False, rvy=rvy)
    kvy.registerReplyRoutes(router=rvy.rtr)

    tvy = Tevery(reger=verifier.reger, db=hby.db, local=False)
    tvy.registerReplyRoutes(router=rvy.rtr)
    tc = TeveryCuery(cdb=cdb, reger=reger, cues=tvy.cues)

    parser = parsing.Parser(framed=True, kvy=kvy, tvy=tvy, rvy=rvy, vry=verifier, exc=exc)

    comms = handling.Communicator(hby=hby, hab=hab, cdb=cdb, reger=reger,
                                  auth=auth, hook=hook, timeout=timeout, retry=retry)
    app.add_route("/health", monitoring.HealthEnd(cdb=cdb))

    ending.loadEnds(app, hby=hby, default=hab.pre)

    doers: List[Doer] = [httpServerDoer, comms, tc]
    if direct:
        logger.info("Adding direct mode HTTP listener")
        # reading notifications for received ipex grant exn messages
        doers.extend(handling.loadHandlers(cdb=cdb, hby=hby, notifier=notifier, parser=parser))

        # Set up HTTP endpoint for PUT-ing application/cesr streams to the SallyAgent at '/'
        httpEnd = indirecting.HttpEnd(rxbs=parser.ims, mbx=mbx)
        app.add_route('/', httpEnd)
        agent = VerificationAgent(hab=hab, parser=parser, kvy=kvy, tvy=tvy, rvy=rvy, verifier=verifier, exc=exc, cues=cues)
        doers.append(agent)
    else:
        logger.info("Adding indirect mode mailbox listener")
        mbd = indirecting.MailboxDirector(
            hby=hby, exc=exc, kvy=kvy, tvy=tvy, rvy=rvy, verifier=verifier, rep=rep,
            topics=["/receipt", "/replay", "/multisig", "/credential", "/delegate", "/challenge"])  # topics to listen for messages on
        # reading notifications for received ipex grant exn messages
        doers.extend(handling.loadHandlers(cdb=cdb, hby=hby, notifier=notifier, parser=mbd.parser))
        doers.append(mbd)

    return doers


def clear_escrows(cdb):
    """Clear escrows if the environment variable CLEAR_ESCROWS is set to True."""
    if env_var_to_bool("CLEAR_ESCROWS", True):
        logger.info("Clearing escrows")
        cdb.clearEscrows()


def env_var_to_bool(var_name, default=False):
    """Convert an environment variable to a boolean value."""
    val = os.getenv(var_name, default)
    if isinstance(val, bool):
        return val
    if isinstance(val, str):
        return val.lower() in ["true", "1"]
    return default

def inception_config(name=None, base=None, alias=None, bran=None, incept_file=None, config_dir=None):
    """
    Inception configuration for Sally's identifier prefix.
    Returns a dict of inception arguments for Sally's identifier prefix.

    This uses an object so that it can be merged with the incept file configuration like the
    multicommand args object by using object properties rather than dictionary keys, which is what
    incept.mergeArgsWithFile expects.
    """
    class Object(object):
        pass
    icp_args = Object()
    icp_args.name = name
    icp_args.base = base
    icp_args.alias = alias
    icp_args.bran = bran
    icp_args.file = incept_file if config_dir is None else os.path.join(config_dir, incept_file)
    icp_args.transferable = True
    icp_args.icount = None
    icp_args.wits = []
    icp_args.toad = None
    icp_args.isith = None
    icp_args.ncount = None
    icp_args.nsith = None
    icp_args.delpre = None
    icp_args.est_only = False
    icp_args.data = None
    return incept.mergeArgsWithFile(icp_args).__dict__




# -*- encoding: utf-8 -*-
"""
sally.cli.commands module

"""
import argparse
import logging
import pprint

import falcon
from hio.core import http
from keri import help
from keri.app import directing

from sally import log_name, ogler
from sally.core import handling, httping
from sally.core.monitoring import HealthEnd

logger = ogler.getLogger(log_name)

parser = argparse.ArgumentParser(description='Launch SALLY sample web hook server')
parser.set_defaults(handler=lambda args: launch(args),
                    transferable=True)
parser.add_argument('-p', '--http',
                    action='store',
                    default=9923,
                    help="Port on which to listen for web hook events.  Defaults to 9923")


def launch(args, expire=0.0):
    baseFormatter = logging.Formatter('%(asctime)s [hook] %(levelname)-8s %(message)s')
    baseFormatter.default_msec_format = None
    help.ogler.baseConsoleHandler.setFormatter(baseFormatter)
    help.ogler.level = logging.getLevelName(logging.INFO)
    help.ogler.reopen(name="hook", temp=True, clear=True)

    httpPort = args.http

    app = falcon.App(
        middleware=httping.cors_middleware())
    app.add_route("/", WebhookListener())
    app.add_route("/health", HealthEnd())

    server = http.Server(port=httpPort, app=app)
    httpServerDoer = http.ServerDoer(server=server)

    logger.info(f"Sally Web Hook Sample listening on {httpPort}")
    directing.runController(doers=[httpServerDoer], expire=expire)


class WebhookListener:
    """
    Demonstration endpoint for web hook calls that prints events to stdout and stores a simple presentation cache.
    """
    def __init__(self):
        self.received = dict()

    def on_post(self, req, resp):
        """Receives web hook POST events by printing the credential results to stdout and storing presentations them in memory
        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
        """
        logger.info("** HEADERS **")
        logger.info(pprint.pprint(req.headers))
        logger.info("*************")

        logger.info("**** BODY ****")
        body = req.get_media()
        logger.info(pprint.pprint(body))
        logger.info("**************")

        data = body.get("data", {})
        if not data:
            logger.error("No data in body")
            resp.media = {"error": "No data in body"}
            resp.status = falcon.HTTP_400
            return
        type = self._resolve_type(data["schema"])

        holder = data.get("recipient", "")
        presentation = dict(
            credential=data.get("credential", ""),
            type=type,
            issuer=body.get("actor", ""),
            holder=holder,
            LEI=data.get("LEI", ""),
            personLegalName=data.get("personLegalName", ""),
            officialRole=data.get("officialRole", ""),
        )
        self.received[holder] = presentation
        resp.status = falcon.HTTP_202

    def _resolve_type(self, schema_said):
        """Return human friendly name for schema type"""
        match schema_said:
            case handling.QVI_SCHEMA:
                return "QVI"
            case handling.LE_SCHEMA:
                return "LE"
            case handling.OOR_AUTH_SCHEMA:
                return "OOR Auth"
            case handling.OOR_SCHEMA:
                return "OOR"
            case _:
                raise ValueError(f"Unknown schema type with SAID: {schema_said}")

    def on_get(self, req, resp):
        """
        Tells the presenter if they have presented a credential before.
        Used in testing to determine that a presentation has succeeded.
        """
        holder = req.get_param("holder", required=True)
        if holder in self.received:
            resp.media = self.received[holder]
            resp.status = falcon.HTTP_200
        else:
            resp.media = {"error": f"No credential presented by {holder}"}
            resp.status = falcon.HTTP_404




# -*- encoding: utf-8 -*-
"""
SALLY
sally.core.httping module

HTTP utility
"""
from base64 import urlsafe_b64encode as encodeB64
from collections import namedtuple

import falcon
from http_sfv import Dictionary
from keri.help import helping

DEFAULTHEADERS = ('(created)', '(request-target)')

Inputage = namedtuple("Inputage", "name fields created keyid alg expires nonce context")


def normalize(param):
    return param.strip()

def cors_middleware():
    """Middleware to allow CORS requests from any origin."""
    falcon.CORSMiddleware(
        allow_origins='*',
        allow_credentials='*',
        expose_headers=cesr_headers())

def cesr_headers():
    """CESR HTTP headers to be exposed in CORS requests."""
    return ['cesr-attachment', 'cesr-date', 'content-type']

def siginput(hab, name, method, path, headers, fields, expires=None, nonce=None, alg=None, keyid=None, context=None):
    """ Create an HTTP Signature-Input Header

    Returns:
        header (dict): {'Signature-Input': 'value'} where value is RFC8941 compliant
        (Structured Field Values for HTTP) formatted str of Signature Input group.
        sigers (Unqualified): unqualified base64 encoded signature

    """
    items = []
    ifields = []

    # Create Signature Base, start with the fields and
    for field in fields:
        if field.startswith("@"):
            if field == "@method":
                items.append(f'"{field}": {method}')
                ifields.append(field)
            elif field == "@path":
                items.append(f'"{field}": {path}')
                ifields.append(field)

        else:
            field = field.lower()
            if field not in headers:
                continue

            ifields.append(field)
            value = normalize(headers[field])
            items.append(f'"{field}": {value}')

    sid = Dictionary()
    sid[name] = ifields
    now = helping.nowUTC()
    sid[name].params['created'] = int(now.timestamp())

    values = [f"({' '.join(ifields)})", f"created={int(now.timestamp())}"]
    if expires is not None:
        values.append(f"expires={expires}")
        sid[name].params['expires'] = expires
    if nonce is not None:
        values.append(f"nonce={nonce}")
        sid[name].params['nonce'] = nonce
    if keyid is not None:
        values.append(f"keyid={keyid}")
        sid[name].params['keyid'] = keyid
    if context is not None:
        values.append(f"context={context}")
        sid[name].params['context'] = context
    if alg is not None:
        values.append(f"alg={alg}")
        sid[name].params['alg'] = alg

    params = ';'.join(values)

    items.append(f'"@signature-params: {params}"')
    ser = "\n".join(items).encode("utf-8")
    sigers = hab.sign(ser=ser,
                      verfers=hab.kever.verfers,
                      indexed=False)

    unq = Unqualified(raw=sigers[0].raw)

    return {'Signature-Input': f"{str(sid)}"}, unq  # join all signature input value strs


def desiginput(value):
    """ Verify the signature header based on values as identified in signature-input header

    Parameters:
        value (Request): falcon request object

    Returns:

    """
    sid = Dictionary()
    sid.parse(value)

    siginputs = []
    for name, svfields in sid.items():
        fields = [i.value for i in svfields]
        if "created" not in svfields.params:
            raise ValueError("missing required `created` field from signature input")
        created = svfields.params["created"]
        if "expires" in svfields.params:
            expires = svfields.params["expires"]
        else:
            expires = None
        if "nonce" in svfields.params:
            nonce = svfields.params["nonce"]
        else:
            nonce = None
        if "alg" in svfields.params:
            alg = svfields.params["alg"]
        else:
            alg = None
        if "keyid" in svfields.params:
            keyid = svfields.params["keyid"]
        else:
            keyid = None
        if "context" in svfields.params:
            context = svfields.params["context"]
        else:
            context = None

        siginputs.append(Inputage(name=name, fields=fields, created=created, expires=expires, nonce=nonce, alg=alg,
                                  keyid=keyid, context=context))
    return siginputs


class Unqualified:

    def __init__(self, raw):
        self._raw = raw

    @property
    def raw(self):
        return self._raw

    @property
    def qb64(self):
        return self.qb64b.decode("utf-8")

    @property
    def qb64b(self):
        return encodeB64(self.raw)



# -*- encoding: utf-8 -*-
"""
SALLY
sally.core.basing module

Database support
"""
from keri.core import coring, serdering
from keri.db import dbing, subing

from sally import log_name, ogler

logger = ogler.getLogger(log_name)

class CueBaser(dbing.LMDBer):
    """
    Noter stores Notifications generated by the agent that are
    intended to be read and dismissed by the controller of the agent.

    """
    TailDirPath = "sally/db"
    AltTailDirPath = ".sally/db"
    TempPrefix = "sally_db_"

    def __init__(self, name="cb", headDirPath=None, reopen=True, **kwa):
        """

        Parameters:
            headDirPath:
            perm:
            reopen:
            kwa:
        """
        self.snd = None

        self.iss = None
        self.rev = None

        self.recv = None
        self.revk = None

        self.ack = None

        super(CueBaser, self).__init__(name=name, headDirPath=headDirPath, reopen=reopen, **kwa)

    def reopen(self, **kwa):
        """

        :param kwa:
        :return:
        """
        super(CueBaser, self).reopen(**kwa)

        # Database of senders of the presentation or revocation messages
        self.snd = subing.CesrSuber(db=self, subkey='snd.', klas=coring.Prefixer)

        # presentations that are waiting for the credential to be received and parsed
        self.iss = subing.CesrSuber(db=self, subkey='iss.', klas=coring.Dater)
        # revocations that are waiting for the TEL event to be received and processed
        self.rev = subing.CesrSuber(db=self, subkey='rev.', klas=coring.Dater)

        # presentations with resolved credentials that need to be sent to the hook
        self.recv = subing.SerderSuber(db=self, subkey="recv", klas=serdering.SerderACDC)
        # revocations whose TEL rev event has been resolved that need to be sent to the hook
        self.revk = subing.SerderSuber(db=self, subkey="revk", klas=serdering.SerderACDC)

        # presentations that have been sent to the hook that need to be ack'ed
        self.ack = subing.SerderSuber(db=self, subkey="ack", klas=serdering.SerderACDC)

        return self.env

    def clearEscrows(self):
        """
        Clear all credential escrows. Useful in testing to avoid many unneeded log messages or force reprocessing of presentations.
        """
        self.iss.trim()
        self.rev.trim()
        self.recv.trim()
        self.revk.trim()
        self.ack.trim()
        logger.info("Cleared iss and rev escrows")

    def getCounts(self):
        """
        Get counts of each database for metrics monitoring
        """
        snd  = len([said for (said,), prefixer in self.snd.getItemIter()])
        iss  = len([said for (said,), dater in self.iss.getItemIter()])
        rev  = len([said for (said,), dater in self.rev.getItemIter()])

        recv = len([said for (said, dater_qb64), creder in self.recv.getItemIter()])
        revk = len([said for (said, dater_qb64), creder in self.revk.getItemIter()])
        ack  = len([said for (said,), creder in self.ack.getItemIter()])
        return {
            'senders': snd,
            'iss': iss,
            'rev': rev,
            'recv': recv,
            'revk': revk,
            'ack': ack
        }


from hio.base import doing
from hio.help import decking
from keri.core import coring

from sally import ogler, log_name

logger = ogler.getLogger(log_name)


class TeveryCuery(doing.Doer):
    """
    Processes credential revocation cues and records when a given credential was revoked in the local cue database (CueBaser).
    """

    def __init__(self, cdb, reger, cues=None, **kwa):
        """
        Parameters:
            cdb (CueBaser): instance of CueBaser database
            reger (Reger): Stores ACDC / TEL events
            cues (Deck): collection of events (cue) to process
        """
        self.cdb = cdb
        self.reger = reger
        self.cues = cues if cues is not None else decking.Deck()

        super(TeveryCuery, self).__init__(**kwa)

    def do(self, tymth, *, tock=0.0, **opts):
        """
        Iterates over the cues Deck and processes each ACDC revocation cue to save the ACDC sender and revocation time in the cue database.

        Inherited Parameters:
            tymth (function): closure for read only injection of cycle time in seconds
            tock (float): cycle time in seconds
        """
        self.wind(tymth)  # Set the clock with cycle time function
        self.tock = tock  # Set the cycle time of this doer
        yield self.tock   # allow priming of this doer to advance to this line

        while True:
            while self.cues:
                cue = self.cues.popleft()
                if cue['kin'] == "revoked":
                    serder = cue["serder"]
                    said = serder.ked["i"]
                    creder = self.reger.creds.get(said)
                    if creder is None:
                        logger.error(f"revocation received for unknown credential {said}")

                    prefixer = coring.Prefixer(qb64=creder.issuer)
                    saider = coring.Saider(qb64=said)
                    now = coring.Dater()

                    self.cdb.snd.pin(keys=(saider.qb64,), val=prefixer)
                    self.cdb.rev.pin(keys=(saider.qb64,), val=now)

                yield self.tock

            yield self.tock



import falcon
import sally
from keri.help import nowIso8601


class HealthEnd:
    """
    Basic health check endpoint including a health message, Sally version, and operational metrics
    """

    def __init__(self, cdb = None):
        """
        Adds the CueBaser to allow getting metric counts.
        Defaults to none in case used via demo webhook
        """
        self.cdb = cdb


    def on_get(self, req, resp):
        counts = self.cdb.getCounts() if self.cdb else {}
        resp.status = falcon.HTTP_OK
        resp.media = {
            "message": f"Health is okay. Time is {nowIso8601()}",
            "version": f"{sally.__version__}",
            "counts": counts
        }


# -*- encoding: utf-8 -*-

"""
SALLY
sally package

"""

__version__ = '1.0.3'  # also change in setup.py

import logging

from hio.help import ogling

from sally.app.logs import TruncatedFormatter

log_name = 'sally'  # name of this project that shows up in log messages
log_format_str = f'%(asctime)s [{log_name}] %(levelname)-8s %(module)s.%(funcName)s-%(lineno)s %(message)s'

ogler = ogling.initOgler(prefix=log_name, syslogged=False)
ogler.level = logging.INFO

formatter = TruncatedFormatter(log_format_str)
formatter.default_msec_format = None
logHandler = logging.StreamHandler()
logHandler.setFormatter(formatter)
ogler.baseFormatter = formatter
ogler.baseConsoleHandler = logHandler
ogler.baseConsoleHandler.setFormatter(formatter)
ogler.reopen(name=log_name, temp=True, clear=True)

def set_log_level(loglevel, logger):
    """Set the log level for the logger."""
    ogler.level = logging.getLevelName(loglevel.upper())
    logger.setLevel(ogler.level)


from hio.base import doing
from hio.help import decking
from sally import ogler, log_name


logger = ogler.getLogger(log_name)


class VerificationAgent(doing.DoDoer):
    """
    Doer for running the reporting agent in direct HTTP mode rather than indirect mode.
    Direct mode is used when presenting directly to the reporting agent after resolving the reporting agent OOBI as a Controller OOBI.
    Indirect mode is used when presenting to the reporting agent via a mailbox whether from a witness or a mailbox agent.
    """

    def __init__(self, hab, parser, kvy, tvy, rvy, verifier, exc, cues=None, **opts):
        """
        Initializes the ReportingAgent with an identifier (Hab), parser, KEL, TEL, and Exchange message processor
        so that it can process incoming credential presentations.

        Parameters:
            hab (Hab): The identifier (Hab) for the reporting agent.
            parser (Parser): The message parser to handle incoming messages.
            kvy (Kevery): The KEL message processor for handling key events.
            tvy (Tevery): The TEL message processor for handling registry and credential events.
            rvy (Revery): The Reply message processor for handling location, endrole, and other reply of messages.
            verifier (Verifier): The Credential processor for handling credential escrows.
            exc (Exchanger): The Exchanger for managing exchange messages.
            cues (Deck, optional): A data buffer for inter-component communication cues. Defaults to an empty deck
        """
        self.hab = hab
        self.parser = parser
        self.kvy = kvy
        self.tvy = tvy
        self.rvy = rvy
        self.verifier = verifier
        self.exc = exc
        self.cues = cues if cues is not None else decking.Deck()
        doers = [doing.doify(self.msgDo), doing.doify(self.escrowDo)]
        super().__init__(doers=doers, **opts)

    def msgDo(self, tymth=None, tock=0.0):
        """
        Processes incoming messages from the parser which triggers the KEL, TEL, Router, and Exchange
        message processor to process credential presentations.
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if self.parser.ims:
            logger.debug(f"ReportingAgent received:\n%s\n...\n", self.parser.ims[:1024])
        done = yield from self.parser.parsator(local=True)
        return done

    def escrowDo(self, tymth=None, tock=0.0):
        """
        Processes KEL, TEL, Router, and Exchange message processor escrows.
        This ensures that each component processes the messages parsed from the HttpEnd.
        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            self.kvy.processEscrows()
            self.rvy.processEscrowReply()
            if self.tvy is not None:
                self.tvy.processEscrows()
            if self.verifier is not None:
                self.verifier.processEscrows()
            self.exc.processEscrow()

            yield


# -*- encoding: utf-8 -*-
"""
SALLY
sally.core.handling module

EXN Message handling
"""
import datetime
import json
from base64 import urlsafe_b64encode as encodeB64
from typing import List
from urllib import parse

from hio.base import doing, Doer
from hio.core import http
from hio.help import Hict
from keri import kering
from keri.core import coring
from keri.peer import exchanging
from keri.end import ending
from keri.help import helping

from sally import ogler, log_name
from sally.core import httping

logger = ogler.getLogger(log_name)

# vLEI ACDC schema SAIDs
QVI_SCHEMA = "EBfdlu8R27Fbx-ehrqwImnK-8Cm79sqbAQ4MmvEAYqao"
LE_SCHEMA = "ENPXp1vQzRF6JwIuS-mp2U8Uf1MoADoP_GqQ62VsDZWY"
OOR_AUTH_SCHEMA = "EKA57bKBKxr_kN7iN5i7lMUxpMG-s19dRcmov1iDxz-E"
OOR_SCHEMA = "EBNaNu-M9P5cgrnfl2Fvymy4E_jvxxyjb70PRtiANlJy"

# makes log messages nicer
type_to_name = {
    QVI_SCHEMA: "QVI",
    LE_SCHEMA: "LE",
    OOR_AUTH_SCHEMA: "OOR Auth",
    OOR_SCHEMA: "OOR"
}


def loadHandlers(cdb, hby, notifier, parser) -> List[Doer]:
    """
    Returns an array of Doers that are handlers for the peer-to-peer exchange messages.
    Sally only uses the notification handler for ACDC presentations.

    Parameters:
        cdb (CueBaser): communication escrow database environment
        hby (Habery): identifier database environment (master keystore)
        notifier (Notifier): Notifications
        parser (Parser): to parse and process each message referred to in an EXN message
    """
    return [PresentationProofHandler(cdb=cdb, hby=hby, notifier=notifier, parser=parser)]


class PresentationProofHandler(doing.Doer):
    """
    Processor for responding to peer-to-peer (exn) notification messages of IPEX Grant presentation proofs.

    """

    def __init__(self, cdb, hby, notifier, parser, **kwa):
        """ Initialize instance

        Parameters:
            cdb (CueBaser): communication escrow database environment
            notifier(Notifier): to read notifications to processes exns
            **kwa (dict): keyword arguments passes to super Doer

        """
        self.cdb = cdb
        self.hby = hby
        self.notifier = notifier
        self.parser = parser
        super(PresentationProofHandler, self).__init__()

    def processNotes(self):
        """
        Handles incoming IPEX Grant presentations by processing the notification queue that is populated when IPEX Grant events occur, which is what happens when a cred

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.

        The payload of the notification exn message is expected to have the following format:
        {
            "r": "/exn/ipex/grant",
            "d": "SAID of the credential"
        }
        The IPEX Grant EXN is then loaded from the Exchanger database.
        The EXN is expected to have the following format:
        {
            "e": { // embedded data in an exchange message.
                "anc": // SerderKERI: anchoring interaction event of the issuance event. Has a reference to the issuance
                {
                    "v":"KERI10JSON00013a_",                                     // CESR version string of this event
                    "t":"ixn",                                                   // KEL Event type
                    "d":"EGhSHKIV5-nkeirdkqzqsvmeF1FXw_yH8NvPSAY1Rgyd",          // SAID digest of the anchoring interaction event
                    "i":"EMl4RhuR_JxpiMd1N8DEJEhTxM3Ovvn9Xya8AN-tiUbl",          // Identifier of the issuer
                    "s":"2",                                                     // Sequence number in the issuer's KEL of the anchoring interaction event.
                    "p":"ED1kkh5_ECYriK-j2gSv6Zjr5way88XVhwRCxk5zoTRG",          // The previous event's digest
                    "a":[ // The anchoring interaction event's seal. Anchors both the issuance TEL event below and the associated ACDC below to the issuer's KEL.
                        {
                            "i":"EElymNmgs1u0mSaoCeOtSsNOROLuqOz103V3-4E-ClXH",  // SAID digest of the ACDC
                            "s": "0",                                            // Sequence number of the TEL event for the TEL associated with the ACDC anchored in this event.
                            "d":"ECUw7AdWEE3fvr7dgbFDXj0CEZuJTTa_H8-iLLAmIUPO"   // SAID digest of the TEL event for the TEL associated with the ACDC anchored in this event.
                        }
                    ]
                },
                "iss": // SerderKERI: registry issuance TEL event body. Has a reference to the ACDC credential.
                {
                    "v":"KERI10JSON0000ed_",                             // CESR version string of this event
                    "t":"iss",                                           // TEL Event type
                    "d":"ECUw7AdWEE3fvr7dgbFDXj0CEZuJTTa_H8-iLLAmIUPO",  // SAID digest of the issuance TEL event.
                    "i":"EElymNmgs1u0mSaoCeOtSsNOROLuqOz103V3-4E-ClXH",  // SAID digest of the ACDC. This refers to the ACDC below.
                    "s":"0",                                             // Sequence number of the TEL issuance event
                    "ri":"EB-u4VAF7A7_GR8PXJoAVHv5X9vjtXew8Yo6Z3w9mQUQ", // Registry identifier of the issuer
                    "dt":"2021-06-27T21:26:21.233257+00:00"              // The datetime the issuance event was created
                },
                "acdc":  // SerderACDC: credential body. The iss event has a reference to this event
                {
                    "v":"ACDC10JSON000197_",                              // CESR version string of this ACDC. indicates serialization type and protocol type
                    "d":"EElymNmgs1u0mSaoCeOtSsNOROLuqOz103V3-4E-ClXH",   // SAID digest of the ACDC
                    "i":"EMl4RhuR_JxpiMd1N8DEJEhTxM3Ovvn9Xya8AN-tiUbl",   // Issuer identifier of the credential
                    "ri":"EB-u4VAF7A7_GR8PXJoAVHv5X9vjtXew8Yo6Z3w9mQUQ",  // Registry identifier that issued this credential
                    "s":"EMQWEcCnVRk1hatTNyK3sIykYSrrFvafX3bHQ9Gkk1kC",   // Schema identifier of the credential
                    "a": { // The credential's attributes; the data the credential contains
                        "d":"EO9_6NattzsFiO8Fw1cxjYmDjOsKKSbootn-wXn9S3iB",  // SAID digest of the attributes section
                        "dt":"2021-06-27T21:26:21.233257+00:00",             // The datetime the credential was created
                        "i":"EMl4RhuR_JxpiMd1N8DEJEhTxM3Ovvn9Xya8AN-tiUbl",  // The identifier of the credential's subject (issuee)
                        "LEI":"254900OPPU84GM83MG36"                         // The Legal Entity Identifier of the credential's subject. This is the data for this credential, a QVI-like credential.
                    }
                }
            }
        }
        """
        for keys, notice in self.notifier.noter.notes.getItemIter():
            logger.info(f"Processing notice {notice.pretty()}")
            attrs = notice.attrs
            route = attrs['r']

            if route == '/exn/ipex/grant':
                # said of grant message
                said = attrs['d']
                exn, pathed = exchanging.cloneMessage(self.hby, said=said)
                embeds = exn.ked['e']

                for label in ("anc", "iss", "acdc"):
                    ked = embeds[label]
                    sadder = coring.Sadder(ked=ked)
                    ims = bytearray(sadder.raw) + pathed[label]
                    self.parser.parseOne(ims=ims)

                acdc = embeds["acdc"]
                said = acdc['d']

                sender = acdc['i']
                prefixer = coring.Prefixer(qb64=sender)

                self.cdb.snd.pin(keys=(said,), val=prefixer)
                self.cdb.iss.pin(keys=(said,), val=coring.Dater())

            # deleting wether its a grant or not, since we only process grant
            self.notifier.noter.notes.rem(keys=keys)

        return False

    def recur(self, tyme):
        """
        On each iteration process exchange (exn) notifications of IPEX Grant presentation notifications.
        """
        self.processNotes()
        return False  # Loop infinitely - long-running Doer task


class Communicator(doing.DoDoer):
    """
    Communicator is responsible for communicating to the webhook the receipt and successful
    verification of credential presentation and revocation messages from external third parties via
    an HTTP API call to the configured webhook URL.
    """

    def __init__(self, hby, hab, cdb, reger, auth, hook, timeout=10, retry=3.0):
        """
        Create a communicator capable of persistent processing of messages and performing
        web hook calls.

        Parameters:
            hby (Habery): identifier database environment
            hab (Hab): identifier environment of this Communicator.  Used to sign hook calls
            cdb (CueBaser): communication escrow database environment
            reger (Reger): credential registry and database
            auth (str): AID of external authority for contacts and credentials
            hook (str): web hook to call in response to presentations and revocations
            timeout (int): escrow timeout (in minutes) for events not delivered to upstream web hook
            retry (float): retry delay (in seconds) for failed web hook attempts
        """
        self.hby = hby
        self.hab = hab
        self.cdb = cdb
        self.reger = reger
        self.hook = hook
        self.auth = auth
        self.timeout = timeout
        self.retry = retry
        self.clients = dict()

        super(Communicator, self).__init__(doers=[doing.doify(self.escrowDo)])

    def processPresentations(self):
        """
        Validate presentations move them to the "received" key/value area if its credential chain
        validates and the credential is not revoked. Otherwise, remove the presentation from the escrow.
        """

        for (said,), dater in self.cdb.iss.getItemIter():
            # cancel presentations that have been around longer than timeout
            now = helping.nowUTC()
            logger.info(f"looking for credential {said}")
            if now - dater.datetime > datetime.timedelta(minutes=self.timeout):
                logger.info(f"presentation {said} timed out")
                self.cdb.iss.rem(keys=(said,))
                continue

            if self.reger.saved.get(keys=(said,)) is not None:
                creder = self.reger.creds.get(keys=(said,))
                try:
                    regk = creder.regi
                    state = self.reger.tevers[regk].vcState(creder.said)
                    if state is None or state.et not in (kering.Ilks.iss, kering.Ilks.bis):
                        raise kering.ValidationError(f"revoked credential {creder.said} being presented")
                    if creder.schema == QVI_SCHEMA:
                        self.validateQualifiedvLEIIssuer(creder)
                    elif creder.schema == LE_SCHEMA:
                        self.validateLegalEntity(creder)
                    elif creder.schema == OOR_SCHEMA:
                        self.validateOfficialRole(creder)
                    else:
                        raise kering.ValidationError(f"credential {creder.said} is of unsupported schema"
                                                     f" {creder.schema} from issuer {creder.issuer}")
                except kering.ValidationError as ex:
                    logger.error(f"Credential {creder.said} from issuer {creder.issuer} failed validation: {ex}")
                else:
                    logger.info(f"Credential {said} from issuer {creder.issuer} passed validation")
                    self.cdb.recv.pin(keys=(said, dater.qb64), val=creder)
                finally:
                    logger.info(f"Removing presentation {said} from escrow after successful validation")
                    self.cdb.iss.rem(keys=(said,))

    def processRevocations(self):
        """
        Ensure revocation CESR data is fully received before moving it to the "revoked to be processed" key/value area.
        """

        for (said,), dater in self.cdb.rev.getItemIter():

            # cancel revocations that have been around longer than timeout
            now = helping.nowUTC()
            if now - dater.datetime > datetime.timedelta(minutes=self.timeout):
                self.cdb.rev.rem(keys=(said,))
                continue

            creder = self.reger.creds.get(keys=(said,))
            if creder is None:  # received revocation before credential.  probably an error but let it timeout
                continue

            regk = creder.regi
            state = self.reger.tevers[regk].vcState(creder.said)
            if state is None:  # received revocation before status.  probably an error but let it timeout
                continue

            elif state.et in (kering.Ilks.iss, kering.Ilks.bis):  # haven't received revocation event yet
                continue

            elif state.et in (kering.Ilks.rev, kering.Ilks.brv):  # revoked
                self.cdb.rev.rem(keys=(said,))
                self.cdb.revk.pin(keys=(said, dater.qb64), val=creder)

    def processReceived(self, db, action):
        """
        Prepare the appropriate payload for issuances or revocations based on schema type and send
        the payload in a request to the webhook URL.
        """

        for (said, dates), creder in db.getItemIter():
            if said not in self.clients:
                resource = creder.schema
                actor = creder.issuer
                if action == "iss":  # presentation of issued credential
                    if creder.schema == QVI_SCHEMA:
                        data = self.qviPayload(creder)
                    elif creder.schema == LE_SCHEMA:
                        data = self.entityPayload(creder)
                    elif creder.schema == OOR_SCHEMA:
                        data = self.roleCredentialPayload(self.reger, creder)
                    else:
                        logger.error(f"invalid credential with schema {creder.schema} said {creder.said} issuer {creder.issuer}")
                        raise kering.ValidationError("this will never happen because all credentials that get here are"
                                                     " valid")
                else:  # revocation of credential
                    data = self.revokePayload(creder)

                logger.info(f"Sending action {action} of {type_to_name[creder.schema]} to {self.hook} with SAID {said}")
                logger.info(f"Payload: \n{json.dumps(data, indent=1)}\n")

                self.request(creder.said, resource, action, actor, data)
                continue

            (client, clientDoer) = self.clients[said]
            if client.responses:
                response = client.responses.popleft()
                self.remove([clientDoer])
                client.close()
                del self.clients[said]

                if 200 <= response["status"] < 300:
                    db.rem(keys=(said, dates))
                    self.cdb.ack.pin(keys=(said,), val=creder)
                else:
                    dater = coring.Dater(qb64=dates)
                    now = helping.nowUTC()
                    if now - dater.datetime > datetime.timedelta(minutes=self.timeout):
                        db.rem(keys=(said, dates))

    def processAcks(self):
        """Once a webhook request is acknowledged then remove it from the ack queue."""
        for (said,), creder in self.cdb.ack.getItemIter():
            # TODO: generate EXN ack message with credential information
            logger.info(f"ACK for credential {said} will be sent to {creder.issuer}")
            self.cdb.ack.rem(keys=(said,))

    def escrowDo(self, tymth, tock=1.0):
        """ Process escrows of comms pipeline

        Steps involve:
           1. Sending local event with sig to other participants
           2. Waiting for signature threshold to be met.
           3. If elected and delegated identifier, send complete event to delegator
           4. If delegated, wait for delegator's anchor
           5. If elected, send event to witnesses and collect receipts.
           6. Otherwise, wait for fully receipted event

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value.  Default to 1.0 to slow down processing

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            try:
                self.processEscrows()
            except Exception as e:
                logger.error(e)

            yield self.retry

    def processEscrows(self):
        """
        Process communication pipelines for presentations, revocations, and webhook HTTP request acknowledgements.

        """
        self.processPresentations()
        self.processRevocations()
        self.processReceived(db=self.cdb.recv, action="iss")
        self.processReceived(db=self.cdb.revk, action="rev")
        self.processAcks()

    def request(self, said, resource, action, actor, data):
        """
        Generate and launch HTTP request to remote webhook URL.
        Adds custom Sally-Resource and Sally-Timestamp headers.

        Parameters:
            said (str): qb64 SAID of credential
            data (dict): serializable body to send with call
            action (str): the action performed on he resource [iss|rev]
            actor (str): qualified b64 AID of sender of the event
            resource (str): the resource type that triggered the event
        """
        purl = parse.urlparse(self.hook)
        client = http.clienting.Client(hostname=purl.hostname, port=purl.port)
        clientDoer = http.clienting.ClientDoer(client=client)
        self.extend([clientDoer])

        body = dict(
            action=action,
            actor=actor,
            data=data
        )

        raw = json.dumps(body).encode("utf-8")
        headers = Hict([
            ("Content-Type", "application/json"),
            ("Content-Length", len(raw)),
            ("Connection", "close"),
            ("Sally-Resource", resource),
            ("Sally-Timestamp", helping.nowIso8601()),
        ])
        path = purl.path or "/"

        keyid = encodeB64(self.hab.kever.serder.verfers[0].raw).decode('utf-8')
        header, unq = httping.siginput(
            self.hab, "sig0", "POST", path, headers,
            fields=[
                "Sally-Resource",
                "@method",
                "@path",
                "Sally-Timestamp"
            ],
            alg="ed25519",
            keyid=keyid
        )

        headers.extend(header)
        signage = ending.Signage(
            markers=dict(sig0=unq), indexed=True, signer=self.hab.pre, ordinal=None, digest=None, kind=None)

        headers.extend(ending.signature([signage]))

        client.request(
            method='POST',
            path=path,
            qargs=parse.parse_qs(purl.query),
            headers=headers,
            body=raw
        )

        self.clients[said] = (client, clientDoer)

    def validateQualifiedvLEIIssuer(self, creder):
        """ Validate issuer of QVI against known valid issuer

        Parameters:
            creder (Creder): QVI credential to validate

        Raises:
            ValidationError: If credential was not issued from known valid issuer
        """
        if creder.schema != QVI_SCHEMA:
            raise kering.ValidationError(f"invalid schema {creder.schema} for QVI credential {creder.said}")

        if not creder.issuer == self.auth:
            logger.info("Creder has an issue: %s", creder.said)
            logger.debug("Creder Body:\n%s\n", creder.pretty())
            raise kering.ValidationError(f"QVI credential not issued by known valid issuer. Expected {self.auth} found {creder.issuer}")

    def validateLegalEntity(self, creder):
        """Validate schema of LE credential and QVI chain"""
        if creder.schema != LE_SCHEMA:
            raise kering.ValidationError(f"invalid schema {creder.schema} for LE credential {creder.said}")

        self.validateQVIChain(creder)

    def validateOfficialRoleAuth(self, creder):
        """Validate schema of OOR Auth credential and the LE chain"""
        if creder.schema != OOR_AUTH_SCHEMA:
            raise kering.ValidationError(f"invalid schema {creder.schema} for OOR credential {creder.said}")

        if creder is None or creder.edge is None:
            raise kering.ValidationError(f"OOR Auth credential does not have expected 'le' edge")
        edges = creder.edge
        lesaid = edges["le"]["n"]
        le = self.reger.creds.get(lesaid)
        if le is None:
            raise kering.ValidationError(f"LE credential {lesaid} not found for AUTH credential {creder.said}")

        self.validateLegalEntity(le)

    def validateOfficialRole(self, creder):
        """Validate OOR schema, the OOR Auth chain, and that the data attributes from the OOR Auth match the OOR credential data"""
        if creder.schema != OOR_SCHEMA:
            raise kering.ValidationError(f"invalid schema {creder.schema} for OOR credential {creder.said}")

        if creder is None or creder.edge is None:
            raise kering.ValidationError(f"OOR credential does not have expected 'auth' edge")
        edges = creder.edge
        asaid = edges["auth"]["n"]
        auth = self.reger.creds.get(asaid)
        if auth is None:
            logger.error(f"AUTH credential {asaid} not found for OOR credential {creder.said}")
            raise kering.ValidationError(f"AUTH credential {asaid} not found for OOR credential {creder.said}")

        if auth.sad["a"]["AID"] != creder.attrib["i"]:
            raise kering.ValidationError(f"invalid issuee {creder.attrib['i']}  doesnt match AUTH value of "
                                         f"{auth.sad['a']['AID']} for OOR " f"credential {creder.said}")

        if auth.sad["a"]["personLegalName"] != creder.attrib["personLegalName"]:
            raise kering.ValidationError(f"invalid personLegalNAme {creder.attrib['personLegalName']} for OOR "
                                         f"credential {creder.said}")

        if auth.sad["a"]["officialRole"] != creder.attrib["officialRole"]:
            raise kering.ValidationError(f"invalid role {creder.attrib['officialRole']} for OOR credential"
                                         f" {creder.said}")

        self.validateOfficialRoleAuth(auth)

    def validateQVIChain(self, creder):
        """Validate that the LE credential has the QVI edge and the QVI chain is valid"""
        if creder is None or creder.edge is None:
            raise kering.ValidationError(f"LE credential does not have expected 'qvi' edge")
        edges = creder.edge
        qsaid = edges["qvi"]["n"]
        qcreder = self.reger.creds.get(qsaid)
        if qcreder is None:
            raise kering.ValidationError(f"QVI credential {qsaid} not found for credential {creder.said}")

        try:
            self.validateQualifiedvLEIIssuer(qcreder)
        except kering.ValidationError as ex:
            logger.info("QVI credential %s failed QVI validation: %s", qsaid, ex)
            logger.debug("QVI credential body:\n%s\n", qcreder.pretty())
            raise ex

    @staticmethod
    def qviPayload(creder):
        """Creates a QVI credential payload to send to the webhook"""
        a = creder.sad["a"]
        data = dict(
            type=type_to_name[creder.schema],
            schema=creder.schema,
            issuer=creder.issuer,
            issueTimestamp=a["dt"],
            credential=creder.said,
            recipient=a["i"],
            LEI=a["LEI"]
        )

        return data

    @staticmethod
    def entityPayload(creder):
        """Creates a legal entity payload to send to the webhook"""
        a = creder.sad["a"]
        if creder is None or creder.edge is None:
            raise kering.ValidationError(f"LE credential does not have expected 'qvi' edge")
        edges = creder.edge
        qsaid = edges["qvi"]["n"]
        data = dict(
            type=type_to_name[creder.schema],
            schema=creder.schema,
            issuer=creder.issuer,
            issueTimestamp=a["dt"],
            credential=creder.said,
            recipient=a["i"],
            qviCredential=qsaid,
            LEI=a["LEI"]
        )

        return data

    @staticmethod
    def roleCredentialPayload(reger, creder):
        """Creates an OOR credential payload to send to the webhook"""
        a = creder.sad["a"]
        if creder is None or creder.edge is None:
            raise kering.ValidationError(f"OOR credential does not have expected 'auth' edge")
        edges = creder.edge
        asaid = edges["auth"]["n"]

        auth = reger.creds.get(asaid)
        if auth is None or auth.edge is None:
            raise kering.ValidationError(f"OOR credential does not have expected 'le' edge")
        aedges = auth.edge
        lesaid = aedges["le"]["n"]
        qvi = reger.creds.get(lesaid)
        if qvi is None or qvi.edge is None:
            raise kering.ValidationError(f"OOR credential does not have expected 'qvi' edge")
        qedges = qvi.edge
        qsaid = qedges["qvi"]["n"]

        data = dict(
            type=type_to_name[creder.schema],
            schema=creder.schema,
            issuer=creder.issuer,
            issueTimestamp=a["dt"],
            credential=creder.said,
            recipient=a["i"],
            authCredential=asaid,
            qviCredential=qsaid,
            legalEntityCredential=lesaid,
            LEI=a["LEI"],
            personLegalName=a["personLegalName"],
            officialRole=a["officialRole"]
        )

        return data

    def revokePayload(self, creder):
        """Creates a revocation payload to send to the webhook"""
        regk = creder.regi
        state = self.reger.tevers[regk].vcState(creder.said)

        data = dict(
            type=type_to_name[creder.schema],
            schema=creder.schema,
            credential=creder.said,
            revocationTimestamp=state.dt
        )

        return data



from typing import Callable

from hio.base import doing
from keri.app import habbing, oobiing

from sally import ogler, log_name

logger = ogler.getLogger(log_name)


class BootstrapRunner(doing.DoDoer):
    def __init__(self, hby: habbing.Habery, tymth: Callable):
        """
        Parameters:
            hby (Habery): the hab in which to create the AID, if needed
            tymth (function): the clock time (scheduler tick rate) to use from the parent Doist
        """
        self.hby = hby
        self.complete = False
        super(BootstrapRunner, self).__init__(tymth=tymth)
        self.extend([OobiResolveBootstrapper(hby, self)])

    def recur(self, tyme, deeds=None):
        if self.complete:
            return True
        super(BootstrapRunner, self).recur(tyme, deeds)
        return False

    def configureAndIncept(self):
        """Configure a keystore by resolving OOBIs in config"""
        self.extend(OobiResolveBootstrapper(self.hby, self))

class OobiResolveBootstrapper(doing.Doer):
    """
    Bootstraps OOBI Resolutions for a Habery and an AID, resolving all OOBIs in the bootstrap configuration file.
    """

    def __init__(self, hby: habbing.Habery, parent: BootstrapRunner, tock=0.0, **kwa):
        self.hby = hby
        self.parent = parent
        super(OobiResolveBootstrapper, self).__init__(tock=tock, **kwa)

    def configure_and_incept(self):
        oobi_count = self.hby.db.oobis.cntAll()
        if oobi_count:
            obi = oobiing.Oobiery(hby=self.hby)
            self.parent.extend(obi.doers)

            while oobi_count > self.hby.db.roobi.cntAll():
                yield 0.25

            for (oobi,), obr in self.hby.db.roobi.getItemIter():
                if obr.state in (oobiing.Result.resolved,):
                    logger.info(f"{oobi} succeeded")
                if obr in (oobiing.Result.failed,):
                    logger.error(f"{oobi} failed")

            self.parent.remove(obi.doers)

        wc = [oobi for (oobi,), _ in self.hby.db.woobi.getItemIter()]
        if len(wc) > 0:
            logger.info(f"\nAuthenticating Well-Knowns...")
            authn = oobiing.Authenticator(hby=self.hby)
            self.parent.extend(authn.doers)

            while True:
                cap = []
                for (_,), wk in self.hby.db.wkas.getItemIter(keys=b''):
                    cap.append(wk.url)

                if set(wc) & set(cap) == set(wc):
                    break

                yield 0.5

            self.parent.remove(authn.doers)

        self.parent.complete = True

    def recur(self, tock=0.0, **opts):
        yield from self.configure_and_incept()
        return True


# -*- encoding: utf-8 -*-
"""
sally.cli.commands module

"""
import argparse
import os

from hio.base import doing
from keri.app import keeping, habbing, configing
from keri.app.cli.common import existing

import sally
from sally import ogler, log_name, set_log_level
from sally.app.bootstrapping import BootstrapRunner
from sally.core import serving

parser = argparse.ArgumentParser(description='Launch Sally vLEI credential presentation receiver service.')
parser.set_defaults(handler=lambda args: launch(args),
                    transferable=True)

parser.add_argument(
    "-d", "--direct", action="store_true", default=False,
    help="Listen for direct-mode messages on HTTP port.  Default is True")
parser.add_argument(
    '-p', '--http', action='store', default=9723,
    help="Port on which to listen for OOBI requests.  Defaults to 9723")
parser.add_argument(
    '-n', '--name', action='store', default="sally",
    help="Name of controller. Default is sally.")
parser.add_argument(
    '-b', '--base', required=False, default="",
    help='additional optional prefix to file location of KERI keystore')
parser.add_argument(
    "-c", "--config-dir", dest="configDir",
    help="directory override for configuration data")
parser.add_argument(
    "-f", '--config-file', dest="configFile", action='store', default=None,
    help="Habery (keystore) configuration filename")
parser.add_argument(
    "-i", "--incept-file", dest="inceptFile", action='store', default=None,
    help="inception configuration filename")
parser.add_argument(
    '-a', '--alias', required=True,
    help='human readable alias for the new identifier prefix')
parser.add_argument(
    '-s', '--salt', required=False,
    help='qualified base64 salt for creating key pairs')
parser.add_argument(
    '--passcode', dest="bran", default=None,
    help='21 character encryption passcode for keystore (is not saved)')
parser.add_argument(
    '-w', '--web-hook', action='store', required=True, default=None,
    help='Webhook address for outbound notifications of credential issuance or revocation')
parser.add_argument(
    '--auth', action="store", required=True,
    help='AID or alias of authority for OOBIs and QVI credential issuer')
parser.add_argument(
    "-r", "--retry-delay", default=10, type=int, action="store",
    help="retry delay (in seconds) for failed web hook attempts")
parser.add_argument(
    "-e", "--escrow-timeout", default=10, type=int, action="store",
    help="timeout (in minutes) for escrowed events that have not been delivered to the web hook.  Defaults to 10")
parser.add_argument(
    "-l", "--loglevel", action="store", required=False, default=os.getenv("SALLY_LOG_LEVEL", "INFO"),
    help="Set log level to DEBUG | INFO | WARNING | ERROR | CRITICAL. Default is CRITICAL")


logger = ogler.getLogger(log_name)

def launch(args, expire=0.0):
    """Launch Sally vLEI credential presentation receiver service"""
    # Logging config
    set_log_level(args.loglevel, logger)

    # Parse arguments
    hook = args.web_hook
    name = args.name
    salt = args.salt
    base = args.base
    bran = args.bran
    try:
        http_port = int(args.http)
    except ValueError:
        raise ValueError(f"Invalid port number: {args.http}. Must be an integer.")
    auth = args.auth
    direct = args.direct

    timeout = args.escrow_timeout
    retry = args.retry_delay

    alias = args.alias
    config_file = args.configFile
    config_dir = args.configDir

    # Inception configuration for when Sally bootstraps itself
    incept_file = args.inceptFile
    incept_args = {
        "name": name, "base": base, "alias": alias, "bran": bran,
        "incept_file": incept_file, "config_dir": config_dir
    }

    hby = init_habery(name=name, base=base, bran=bran, config_file=config_file,
                      config_dir=config_dir, salt=salt)
    hab = serving.incept_if_new(hby, alias, incept_args)

    # setup doers
    doers = serving.setupDoers(hby, hab, alias=alias, http_port=http_port, hook=hook, auth=auth,
                           timeout=timeout, retry=retry, direct=direct, incept_args=incept_args)
    logger.info(f"Sally Server v{sally.__version__} listening on {http_port} with DB version {hby.db.version}")
    serving.run_doers(doers)

def init_habery(name, base, bran, config_file, config_dir, salt):
    """Initialize or reopen the Habery (keystore)"""
    ks = keeping.Keeper(name=name, base=base, temp=False, reopen=True)
    aeid = ks.gbls.get('aeid')

    cf = None
    if aeid is None:
        if config_file is not None:
            cf = configing.Configer(name=config_file, base=base, headDirPath=config_dir,
                                    temp=False, reopen=True, clear=False)
        habery_cfg = dict()
        habery_cfg["salt"] = salt if salt else None # When None causes Habery to randomize salt
        hby = habbing.Habery(name=name, base=base, bran=bran, cf=cf, ks=ks, **habery_cfg)
    else:
        ks.close() # existing.setupHby reopens, so close here to avoid LMDB conflict.
        hby = existing.setupHby(name=name, base=base, bran=bran)

    # Resolve OOBIs for the Habery and the new Hab
    bootstrap_doist = doing.Doist(limit=0.0, tock=0.03125, real=True)
    runner = BootstrapRunner(hby=hby, tymth=bootstrap_doist.tymen())
    bootstrap_doist.do(doers=[runner])
    return hby


import logging


class TruncatedFormatter(logging.Formatter):
    """
    Formats log records with shorter module and function names and a right justified line number for readability.

    Example:
        "2025-09-05 13:59:29 [keria] INFO     serving    .runAgency       -  145 The Agency is loaded and waiting for requests..."
    """

    def __init__(self, fmt=None, datefmt=None, style='%'):
        super().__init__(fmt, datefmt, style)

    def format(self, record):
        """
        Format the log record with truncated module and function names.
        Ignores exceptions and logs an error if formatting fails.
        """
        # Truncate module and funcName to first 'chars' characters
        mod_chars = 10  # number of spaces to truncate to
        fn_chars = 14  # number of spaces to truncate to
        record.module = (record.module[:mod_chars] + ' ' * mod_chars)[:mod_chars]
        record.funcName = (record.funcName[:fn_chars] + ' ' * fn_chars)[:fn_chars]
        record.lineno = str(record.lineno).rjust(5)  # Ensure line number is right-aligned
        try:
            return super().format(record)
        except Exception as e:
            logging.error(f'Error formatting log record: {e}')
            raise e



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
$ python setup.py register sdist upload

First Time register project on pypi
https://pypi.org/manage/projects/


Pypi Release
$ pip3 install twine

$ python3 setup.py sdist
$ twine upload dist/sally-0.0.1.tar.gz

Create release git:
$ git tag -a v0.4.2 -m "bump version"
$ git push --tags
$ git checkout -b release_0.4.2
$ git push --set-upstream origin release_0.4.2
$ git checkout master

Best practices for setup.py and requirements.txt
https://caremad.io/posts/2013/07/setup-vs-requirement/
"""

from glob import glob
from os.path import basename
from os.path import splitext
from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
if (this_directory / "README.md").exists():  # If building inside a container, like in the `images/keria.dockerfile`, this file won't exist and fails the build
    long_description = (this_directory / "README.md").read_text()
else:
    long_description = "vLEI auditing server that responds to credential presentations by signaling via webhooks."

setup(
    name='sally',
    version='1.0.3',  # also change in src/sally/__init__.py
    license='Apache-2.0',
    license_files=('LICENSE',),
    description='vLEI Audit Reporting API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Philip S. Feairheller',
    author_email='pfeairheller@gmail.com',
    url='https://github.com/GLEIF-IT/sally',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: PyPy',
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
    ],
    project_urls={
        'Documentation': 'https://sally.readthedocs.io/',
        'Changelog': 'https://sally.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/WebOfTrust/sally/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='>=3.12.3',
    install_requires=[
        # This version of KERIpy uses the GLEIF-IT/keripy v1.2.8 branch as of the exn message fix
        # 'keri @ git+https://github.com/GLEIF-IT/keripy.git@e881c9522b9bd38c5d5d5ef04f4e231eb902c36a',

        # TODO remove this and uncomment the above dep - only used for editable installs
        'keri==1.2.7',

        'hio==0.6.14',
        'multicommand==1.0.0',
        'blake3==0.4.1',
        'falcon==4.0.2',
        'http_sfv==0.9.9'
    ],
    extras_require={
        'test': ['pytest', 'coverage', 'pytest-mock-server'],
        'docs': ['sphinx', 'sphinx-rtd-theme']
    },
    entry_points={
        'console_scripts': [
            'sally = sally.app.cli.kli:main',
        ]
    },
)



# -*- encoding: utf-8 -*-
"""
sally.app.commands module

"""
import os

import multicommand

from sally.app.cli import commands


def main():
    parser = multicommand.create_parser(commands)
    args = parser.parse_args()

    if not hasattr(args, 'handler'):
        parser.print_help()
        return

    try:
        args.handler(args)
    except Exception as ex:
        if os.getenv('DEBUG_KLI'):
            import traceback
            traceback.print_exc()
        else:
            print(f"ERR: {ex}")
        return -1


if __name__ == "__main__":
    main()
